"""
Some utilities. Also, a more functional API to the execution of graphs.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

import json
import copy
import warnings


def execute_graph_from_context(graph, context, *targets, inplace=False, check_feasibility=True):
    """Execute a graph up to a target given a context.

    Parameters
    ----------
    graph : grapes Graph
        Graph of the computation.
    context : dict
        Dictionary of the initial context of the computation (input).
    targets : strings (or keys in the graph)
        Indicator of what to compute (desired output).
    inplace : bool
        Whether to modify graph and context inplace (default: False).
    check_feasibility : bool
        Whether to check the feasibility of the computation, which slows performance (default: True).

    Returns
    -------
    grapes Graph
        Graph with context updated after computation.
    """
    if check_feasibility:
        feasibility, missing_dependencies = check_feasibility_of_execution(graph, context, *targets, inplace=inplace)
        if feasibility == "unreachable":
            raise ValueError("The requested computation is unfeasible because of the following missing dependencies: " + ", ".join(missing_dependencies))
        elif feasibility == "uncertain":
            warnings.warn("The feasibility of the requested computation is uncertain because of the following missing dependencies: " + ", ".join(missing_dependencies))

    if not inplace:
        graph = copy.deepcopy(graph)
        context = copy.deepcopy(context)

    graph.set_internal_context(context)
    graph.execute_to_targets(*targets)

    return graph


def json_from_graph(graph):
    """Get a JSON string representing the context of a graph.

    Parameters
    ----------
    graph : grapes Graph
        Graph containing the context to convert to JSON.

    Returns
    -------
    str
        JSON string that prettily represents the context of the graph.
    """

    context = graph.get_internal_context(exclude_recipes=True)
    non_serializable_items = {}
    for key, value in context.items():
        try:
            json.dumps(value)
        except:
            non_serializable_items.update({key: str(value)})
    if len(non_serializable_items) > 0:  # We must copy the context, to preserve it, and dump a modified version of it
        res = copy.deepcopy(context)
        res.update(non_serializable_items)
    else:
        res = context
    return json.dumps(res, sort_keys=True, indent=4, separators=(',', ': '))


def context_from_json_file(file_name):
    """
    Load a json file into a dictionary.

    Parameters
    ----------
    file_name: str
        Path to the json file.

    Returns
    dict
        Content of the file as dictionary.
    """
    with open(file_name, encoding='utf-8') as json_file:
        data = json.load(json_file)
    return data


def wrap_graph_with_function(graph, input_keys, *targets, constants={}, input_as_kwargs=True):
    # Copy graph so as not to pollute the original
    operational_graph = copy.deepcopy(graph)
    # Pass all constants to the graph
    operational_graph.update_internal_context(constants)
    # Freeze so that the constants are fixed
    operational_graph.freeze()
    # Unfreeze the input.
    # Note that this has precedence over constants (i.e., if a key is both input and constant, it is treated as input)
    if len(input_keys) > 0:
        operational_graph.unfreeze(*input_keys)
        operational_graph.clear_values(*input_keys)
    # Move as much as possible towards targets
    operational_graph.progress_towards_targets(*targets)
    # Check feasibility
    placeholder_value = 0
    context = {key: placeholder_value for key in input_keys}
    feasibility, missing_dependencies = check_feasibility_of_execution(operational_graph, context, *targets)
    if feasibility == "unreachable":
        raise ValueError("The requested computation is unfeasible because of the following missing dependencies: " + ", ".join(missing_dependencies))
    elif feasibility == "uncertain":
        warnings.warn("The feasibility of the requested computation is uncertain because of the following missing dependencies: " + ", ".join(missing_dependencies))

    if input_as_kwargs:
        def specific_function(**kwargs):
            # Use for loop rather than dict comprehension because it is a more basic operation
            for key in input_keys:
                operational_graph[key] = kwargs[key]
            operational_graph.execute_to_targets(*targets)
            list_of_values = operational_graph.get_list_of_values(targets)
            # Clear values so that the function can be called again
            operational_graph.clear_values()
            if len(list_of_values) == 1:
                return list_of_values[0]
            else:
                return list_of_values
    else:
        input_keys = list(input_keys)

        def specific_function(*args):
            # Use for loop rather than dict comprehension because it is a more basic operation
            for i in range(len(input_keys)):
                operational_graph[input_keys[i]] = args[i]
            operational_graph.execute_to_targets(*targets)
            list_of_values = operational_graph.get_list_of_values(targets)
            # Clear values so that the function can be called again
            operational_graph.clear_values()
            if len(list_of_values) == 1:
                return list_of_values[0]
            else:
                return list_of_values
    return specific_function


def lambdify_graph(graph, input_keys, target):
    graph = copy.deepcopy(graph)
    graph.unfreeze()
    if len(input_keys) > 0:
        graph.clear_values(*input_keys)
    graph.progress_towards_targets(target)
    while not set(graph.get_args(target) + tuple(graph.get_kwargs(target).values())).issubset(set(input_keys)):
        graph.simplify_all_dependencies(target, exclude=input_keys)
    return graph[graph.get_recipe(target)]


def check_feasibility_of_execution(graph, context, *targets, inplace=False):
    if not inplace:
        graph = copy.deepcopy(graph)
        context = copy.deepcopy(context)

    graph.clear_reachabilities()
    graph.set_internal_context(context)
    graph.find_reachability_targets(*targets)
    feasibility = graph.get_worst_reachability(*targets)
    missing_dependencies = set()
    if feasibility in {"unreachable", "uncertain"}:
        for node in graph.nodes:
            if graph.get_topological_generation_index(node) == 0 and graph.has_reachability(node) and graph.get_reachability(node) != "reachable":
                missing_dependencies.add(node)
    return feasibility, missing_dependencies



"""
Tools that allow the visualization of a graph using graphviz. Requires pygraphviz.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

import networkx
from matplotlib.cm import get_cmap
from .. import *


def get_graphviz_digraph(graph, hide_recipes=False, pretty_names=False, include_values=False, color_by_generation=False, colormap="viridis", **attrs):
    # Get a graphviz AGraph
    g = nx.drawing.nx_agraph.to_agraph(graph._nxdg)
    # Add attributes to the AGraph
    g.graph_attr.update(**attrs)
    # Save some values that will be useful later
    max_topological_generation_index = len(graph.get_topological_generations())-1
    cmap = get_cmap(colormap)

    for node_name in g.nodes():
        new_attrs = {}
        node = graph._nxdg.nodes[node_name]

        # Remove recipes if needed, or eliminate attribute of function
        if node["is_recipe"] and hide_recipes:
            g.remove_node(node_name)
            continue
        elif node["is_recipe"] and node["has_value"]:
            new_attrs.update(value="function")

        # Prettify label if required
        label = prettify_label(node_name) if pretty_names else node_name

        # Add values to the label if required
        if include_values and node.attr["has_value"]:
            if isinstance(node.attr["value"], float):
                value_in_label = "{:.2e}".format(node.attr["value"])
            else:
                value_in_label = str(node.attr["value"])
            label += "\n" + value_in_label.partition('\n')[0]
            if value_in_label.find("\n") != -1:
                label += "\n..."
        new_attrs.update(label=label)

        # Manipulate shapes
        if node["is_recipe"]:
            shape = "ellipse"
        elif node["type"] == "conditional":
            shape = "diamond"
        else:
            shape = "box"
        new_attrs.update(shape=shape)

        # Manipulate colors
        if color_by_generation:
            topological_generation_index = graph.get_node_attribute(node_name, "topological_generation_index")
            # The __call__ method of the colormap returns a tuple of rgba values in [0, 1]
            # We call it passing the ratio between the topological_generation_index of this node and the max of the graph
            color_rgba = cmap(topological_generation_index/max_topological_generation_index)
            # Note that graphviz wants colors in hex form
            new_attrs.update(style="filled", fillcolor=hex_string_from_rgba(*color_rgba), fontcolor=hex_string_from_rgba(*best_text_from_background_color(*color_rgba)))

        # Pass these attributes to the actual AGraph
        g.get_node(node_name).attr.update(new_attrs)

        # Handle edge shapes
        special_dependencies = []
        if node["type"] == "standard" and "recipe" in node:
            if not hide_recipes:
                g.get_edge(node["recipe"], node_name).attr.update(arrowhead="dot")
            special_dependencies.append(node["recipe"])
        elif node["type"] == "conditional":
            for condition in node["conditions"]:
                g.get_edge(condition, node_name).attr.update(arrowhead="diamond")
            special_dependencies.extend(node["conditions"])

    # Return the AGraph (no layout is computed yet)
    return g


def prettify_label(name):
    return "".join(c.upper() if ((i > 0 and name[i-1] == "_") or i == 0) else c for i, c in enumerate(name)).replace("_", " ")


def hex_string_from_rgba(r, g, b, a):
    """
    Get a formatted hex string from RGBA values.

    Parameters
    ----------
    r: float
        Red color channel in [0, 1].
    g: float
        Green color channel in [0, 1].
    b: float
        Blue color channel in [0, 1].
    a: float
        Alpha color channel in [0, 1].

    Returns
    -------
    str
        Formatted hex string starting with # and then 8 hex characters (4 bytes).
    """
    # Convert float values in [0, 1] to hex strings of two characters, e.g. 1->ff
    r_hex = f'{int(r*255):02x}'
    g_hex = f'{int(g*255):02x}'
    b_hex = f'{int(b*255):02x}'
    a_hex = f'{int(a*255):02x}'
    return u"#" + r_hex + g_hex + b_hex + a_hex


def best_text_from_background_color(r, g, b, a=1):
    """
    Get the best text color from background.

    Parameters
    ----------
    r: float
        Red color channel in [0, 1].
    g: float
        Green color channel in [0, 1].
    b: float
        Blue color channel in [0, 1].
    a: float
        Alpha color channel in [0, 1] (default: 1). This value is ignored.

    Returns
    -------
    tuple
        Tuple of RGBA values representing pure white (1, 1, 1, 1) or pure black (0, 0, 0, 1) depending on luminance of input color.
    """
    luminance = 0.212*r + 0.701*g + 0.087*b
    return (1, 1, 1, 1) if luminance < 0.5 else (0, 0, 0, 1)



"""
Tests of visualization.

In many tests, the source of a graphviz graph is compared to an expected value stored in a pickle (pkl) file.
To build a new test of this kind, simply insert a line like
expected_sources[name] = gv.string()
just before the assertion. This populates the expected file with the graph produced by the test.
After running the test once (with pytest), remove that line, otherwise the test always passes.
Remember to use LF line endings in the pkl file.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

import pytest
import grapes as gr
import grapes.visualize
import filecmp
import pickle

output_directory = "tests/visualizations"
expected_directory = "tests/expected"


@pytest.fixture(scope="module")
def expected_sources():
    with open(expected_directory + "/expected.pkl", "rb") as f:
        expected = pickle.load(f)
    yield expected
    with open(expected_directory + "/expected.pkl", "wb") as f:
        pickle.dump(expected, f, 0)


def build_graph():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")
    g.finalize_definition()
    return g


def test_simple(expected_sources):
    g = build_graph()
    name = "simple"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]


def test_with_values(expected_sources):
    g = build_graph()
    g.set_internal_context({"a": 1, "b": 2, "f": 12, "op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y})
    name = "with_values"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]


def test_attrs(expected_sources):
    g = build_graph()
    name = "attrs"
    gv = gr.visualize.get_graphviz_digraph(g, rankdir="LR")
    assert gv.string() == expected_sources[name]


def test_no_operations(expected_sources):
    g = build_graph()
    name = "no_operations"
    gv = gr.visualize.get_graphviz_digraph(g, hide_recipes=True)
    assert gv.string() == expected_sources[name]


def test_save_and_render(expected_sources):
    g = build_graph()
    name = "simple"
    gv = gr.visualize.get_graphviz_digraph(g)
    gv.write(output_directory + "/" + name + ".gv")
    assert filecmp.cmp(output_directory + "/" + name + ".gv", expected_directory + "/" + name + ".gv")
    gv.draw(output_directory + "/" + name + ".gv.pdf", format="pdf", prog="dot")
    assert filecmp.cmp(output_directory + "/" + name + ".gv.pdf", expected_directory + "/" + name + ".gv.pdf")


def test_conditional(expected_sources):
    g = gr.Graph()
    g.add_simple_conditional("d", "c", "a", "b")
    name = "conditional"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]


def test_simplify_dependency(expected_sources):
    g = build_graph()
    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)

    name = "presimplification"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]

    g.simplify_dependency("g", "f")
    name = "postsimplification"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]


def test_simplify_all_dependencies(expected_sources):
    g = build_graph()
    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)

    g.simplify_all_dependencies("g")
    name = "postallsimplification"
    gv = gr.visualize.get_graphviz_digraph(g)
    assert gv.string() == expected_sources[name]


def test_color_by_generation(expected_sources):
    g = build_graph()
    gv = gr.visualize.get_graphviz_digraph(g, color_by_generation=True, colormap="plasma")
    name = "color_by_generation"
    assert gv.string() == expected_sources[name]



"""
Some tools to compose functions into one.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

import inspect


def identity_token():
    """
    A trivial token that has the only purpose of being identifiable
    """
    pass


def function_compose(func, subfuncs, func_dependencies, subfuncs_dependencies, func_signature, subfuncs_signatures):
    """
    Compose functions.

    Parameters
    ----------
    func: callable
        External function
    subfuncs: list of callables
        List of internal functions to be composed with func. If identity_token is passed here, it is replaced by the value of the argument.
    func_dependencies: list of hashables
        Names of the arguments of the new function (usually they should correspond to node names in a graph)
    subfuncs_dependencies: list of lists of hashables
        Names of the arguments that are passed to the subfuncs when calling the new function (usually they should correspond to node names in a graph)
    func_signature: list of hashables
        Names of the arguments of the old func
    subfuncs_dependencies: list of lists of hashables
        Names of the arguments of the old subfuncs
    """
    return lambda **kwargs: func(**{func_signature[i]: (subfuncs[i](**{signature_name: kwargs[dependency_name] for signature_name, dependency_name in zip(subfuncs_signatures[i], subfuncs_dependencies[i])}) if subfuncs[i] is not identity_token else kwargs[func_dependencies[i]]) for i in range(len(subfuncs))})


def function_compose_simple(func, subfuncs, func_dependencies, subfuncs_dependencies, func_signature=None, subfuncs_signatures=None):
    """
    Compose functions, without the need to pass signatures, as they are found automatically.

    Parameters
    ----------
    func: callable
        External function
    subfuncs: list of callables
        List of internal functions to be composed with func
    func_dependencies: list of hashables
        Names of the arguments of the new function (usually they should correspond to node names in a graph)
    subfuncs_dependencies: list of lists of hashables
        Names of the arguments that are passed to the subfuncs when calling the new function (usually they should correspond to node names in a graph)
    func_signature: list of hashables
        Names of the arguments of the old func
    subfuncs_dependencies: list of lists of hashables
        Names of the arguments of the old subfuncs
    """
    if func_signature is None:
        if inspect.getfullargspec(func).varargs is not None:
            raise ValueError("Functions with varargs are not supported by Function Composer")
        elif inspect.getfullargspec(func).varargs is None and inspect.getfullargspec(func).varkw is None:  # Well defined spec
            func_signature = list(inspect.signature(func).parameters.keys())
        else:
            func_signature = func_dependencies
    if subfuncs_signatures is None:
        subfuncs_signatures = []
        for index, subfunc in enumerate(subfuncs):
            if inspect.getfullargspec(func).varargs is not None:
                raise ValueError("Functions with varargs are not supported by Function Composer")
            elif inspect.getfullargspec(subfunc).varargs is None and inspect.getfullargspec(subfunc).varkw is None:  # Well defined spec
                this_signature = list(inspect.signature(subfunc).parameters.keys())
            else:
                this_signature = subfuncs_dependencies[index]
            subfuncs_signatures.append(this_signature)
    return function_compose(func, subfuncs, func_dependencies, subfuncs_dependencies, func_signature, subfuncs_signatures)



"""
Tests of core functionality.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

import pytest
import grapes as gr
import warnings


def test_simple():
    g = gr.Graph()
    g.add_step("a")
    g.add_step("b")
    g.add_step("c")
    g.add_step("d")
    g.add_step("op_e")
    g.add_step("op_f")
    g.add_step("op_g")
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "f": 12, "op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}, "g")
    assert res["g"] == -9


def test_simplified_input():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "f": 12, "op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}, "g")
    assert res["g"] == -9


def test_diamond():
    g = gr.Graph()
    g.add_step("a")
    g.add_step("b", "op_b", "a")
    g.add_step("c", "op_c", "b")
    g.add_step("d", "op_d", "b")
    g.add_step("e", "op_e", "c", "d")

    res = gr.execute_graph_from_context(g, {"a": 1, "op_b": lambda x: 2*x, "op_c": lambda x: 2*x, "op_d": lambda x: 2*x, "op_e": lambda x, y: x-y}, "e")
    assert res["e"] == 0


def test_inverted_input():
    # Typically, we proceed from bottom to top
    # Here we test the opposite
    g = gr.Graph()
    g.add_step("c", "op_c", "b")
    g.add_step("b", "op_b", "a")

    res = gr.execute_graph_from_context(g, {"a": 1, "op_b": lambda x: 2*x, "op_c": lambda x: 3*x}, "c")
    assert res["c"] == 6


def test_conditional():
    g = gr.Graph()
    g.add_simple_conditional("d", "c", "a", "b")
    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "c": True}, "d")
    assert res["d"] == res["a"]


def test_compatibility():
    g = gr.Graph()
    g.add_step("c", "op_c", "a", "b")
    h = gr.Graph()
    h.add_step("e", "op_e", "c", "d")
    assert h.is_compatible(g)


def test_incompatibility():
    g = gr.Graph()
    g.add_step("c", "op_c", "a", "b")
    h = gr.Graph()
    h.add_step("c", "op_c2", "a", "d")
    assert not h.is_compatible(g)


def test_merge():
    exp = gr.Graph()
    exp.add_step("c", "op_c", "a", "b")
    exp.add_step("e", "op_e", "c", "d")

    g = gr.Graph()
    g.add_step("c", "op_c", "a", "b")
    h = gr.Graph()
    h.add_step("e", "op_e", "c", "d")
    g.merge(h)

    assert g == exp


def test_merge_and_execute():
    exp = gr.Graph()
    exp.add_step("c", "op_c", "a", "b")
    exp.add_step("e", "op_e", "c", "d")

    g = gr.Graph()
    g.add_step("c", "op_c", "a", "b")
    h = gr.Graph()
    h.add_step("e", "op_e", "c", "d")
    g.merge(h)

    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "d": 4, "op_c": lambda x, y: x+y, "op_e": lambda x, y: x*y}, "e")
    assert res["e"] == 12


def test_kwargs():
    g = gr.Graph()
    g.add_step("c", "op_c", "a", exponent="b")

    def example_exponentiation_func(base, exponent):
        return base**exponent
    res = gr.execute_graph_from_context(g, {"a": 5, "b": 2, "op_c": example_exponentiation_func}, "c")
    assert res["c"] == 25


def test_wrap_with_function():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)
    g.finalize_definition()

    # Get a function a,b,c,d -> g
    f1 = gr.wrap_graph_with_function(g, ["a", "b", "c", "d"], "g", input_as_kwargs=False)
    # Get a function a,b,c,d -> [e,f,g]
    f2 = gr.wrap_graph_with_function(g, ["a", "b", "c", "d"], "e", "f", "g", input_as_kwargs=False)
    assert f1(1, 2, 3, 4) == -9
    assert f2(1, 2, 3, 4) == [3, 12, -9]


def test_lambdify():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)
    g.finalize_definition()

    # Get a function a,b,c,d -> g
    f1 = gr.lambdify_graph(g, ["a", "b", "c", "d"], "g")
    assert f1(a=1, b=2, c=3, d=4) == -9


def test_simplify_dependency():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)

    g.simplify_dependency("g", "f")
    g.finalize_definition()

    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "c": 3, "d": 4}, "g")
    assert res["g"] == -9


def test_simplify_all_dependencies():
    g = gr.Graph()
    g.add_step("e", "op_e", "a", "b")
    g.add_step("f", "op_f", "c", "d")
    g.add_step("g", "op_g", "e", "f")

    operations = {"op_e": lambda x, y: x+y, "op_f": lambda x, y: x*y, "op_g": lambda x, y: x-y}
    g.set_internal_context(operations)

    g.simplify_all_dependencies("g")
    g.finalize_definition()

    res = gr.execute_graph_from_context(g, {"a": 1, "b": 2, "c": 3, "d": 4}, "g")
    assert res["g"] == -9


def test_progress_towards_targets():
    g = gr.Graph()
    g.add_step("b", "op_b", "a")
    g.add_step("f", "op_f", "b", "c", "e")
    g.add_step("e", "op_e", "d")

    context = {"op_b": lambda x: 2*x, "op_e": lambda x: 3*x, "op_f": lambda x, y, z: x + y + z, "a": 5, "d": 4}
    g.set_internal_context(context)
    g.finalize_definition()

    # f cannot be reached because c is not in context, but b and e can be computed
    g.progress_towards_targets("f")
    assert g["b"] == 10
    assert g["e"] == 12


def test_multiple_conditional():
    g = gr.Graph()
    g.add_multiple_conditional("result", ["condition_1", "condition_2", "condition_3"], ["node_1", "node_2", "node_3"])
    context = {
        "condition_1": False,
        "condition_2": True,
        "condition_3": False,
        "node_1": 1,
        "node_2": 2,
        "node_3": 3,
    }
    g.set_internal_context(context)
    g.finalize_definition()

    g.execute_to_targets("result")

    assert g["result"] == 2


def test_multiple_conditional_with_default():
    g = gr.Graph()
    g.add_multiple_conditional("result", ["condition_1", "condition_2", "condition_3"], ["node_1", "node_2", "node_3", "node_default"])
    context = {
        "condition_1": False,
        "condition_2": False,
        "condition_3": False,
        "node_1": 1,
        "node_2": 2,
        "node_3": 3,
        "node_default": 4
    }
    g.set_internal_context(context)
    g.finalize_definition()

    g.execute_to_targets("result")

    assert g["result"] == 4


def test_edit_step():
    g = gr.Graph()
    g.add_step("b", "op_b", "a")
    g.add_step("c", "op_c", "b")
    g.set_internal_context({"a": 1, "op_b": lambda x: 2*x, "op_c": lambda x: 3*x})
    g.finalize_definition()

    g.execute_to_targets("c")
    assert g["b"] == 2
    assert g["c"] == 6

    g.edit_step("b", "new_op_b", "a", "d")
    assert g["b"] == 2  # Value is unchanged
    assert g["c"] == 6  # Value is unchanged

    g.clear_values("b", "c")
    g.update_internal_context({"d": 3, "new_op_b": lambda x, y: x + y})
    g.finalize_definition()

    g.execute_to_targets("c")
    assert g["b"] == 4
    assert g["c"] == 12


def test_remove_step():
    g = gr.Graph()
    g.add_step("b", "op_b", "a")
    g.add_step("c", "op_c", "b")
    g.set_internal_context({"a": 1, "op_b": lambda x: 2*x, "op_c": lambda x: 3*x})
    g.finalize_definition()

    g.remove_step("b")
    with pytest.raises(KeyError):
        g["b"]
    with pytest.raises(ValueError):
        g.remove_step("d")


def test_add_step_quick():
    def example_function_only_positional(a, b):
        return a**b

    def example_function_with_kw_only_args(a, b, *, c):
        return a**b+c

    def example_function_with_no_args():
        return 1

    g = gr.Graph()
    g.add_step_quick("c", example_function_only_positional)
    g.add_step_quick("d", example_function_with_kw_only_args)
    g.add_step_quick("e", example_function_with_no_args)
    g.add_step_quick("f", lambda e: 2*e)
    g.update_internal_context({"a": 2, "b": 3})
    g.finalize_definition()

    g.execute_to_targets("d", "f")  # "c" and "e" should be automatically computed
    assert g["c"] == 8
    assert g["d"] == 16
    assert g["e"] == 1
    assert g["f"] == 2

    def example_function_with_varargs(*args):
        return 1

    with pytest.raises(ValueError):
        g.add_step_quick("g", example_function_with_varargs)
    with pytest.raises(TypeError):
        g.add_step_quick("h", "a non-function object")


def test_topological_generations():
    g = gr.Graph()
    g.add_step("d", "fd", "b", "c")
    g.add_step("b", "fb", "a")
    g.finalize_definition()

    assert g.get_node_attribute("a", "topological_generation_index") == 0
    assert g.get_node_attribute("b", "topological_generation_index") == 1
    assert g.get_node_attribute("c", "topological_generation_index") == 0
    assert g.get_node_attribute("d", "topological_generation_index") == 2
    assert g.get_node_attribute("fb", "topological_generation_index") == 0
    assert g.get_node_attribute("fd", "topological_generation_index") == 0


def test_reachability_simple():
    g = gr.Graph()
    g.add_step("b", "fb", "a")
    g["fb"] = lambda a: a
    g.finalize_definition()

    g.find_reachability_targets("b")
    assert g.get_reachability("b") == "unreachable"
    g.clear_reachabilities()

    g.update_internal_context({"a": 1})
    g.find_reachability_targets("b")
    assert g.get_reachability("b") == "reachable"


def test_reachability_long_graph():
    g = gr.Graph()
    g.add_step_quick("c", lambda b: b)
    g.add_step_quick("b", lambda a: a)
    g.finalize_definition()

    g.find_reachability_targets("b")
    assert g.get_reachability("b") == "unreachable"
    g.clear_reachabilities()

    g.update_internal_context({"a": 1})
    g.find_reachability_targets("b")
    assert g.get_reachability("b") == "reachable"


def test_reachability_conditional_with_true_value():
    g = gr.Graph()
    g.add_simple_conditional("name", "condition", "value_true", "value_false")
    g["condition"] = True
    g.finalize_definition()

    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    g.update_internal_context({"value_true": 1})
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "reachable"


def test_reachability_multiple_conditional_with_true_value():
    g = gr.Graph()
    g.add_multiple_conditional("name", ["ca", "cb"], ["a", "b", "c"])
    g["ca"] = True
    g.finalize_definition()

    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    g.update_internal_context({"a": 1})
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "reachable"


def test_conditional_no_conditions_defined():
    g = gr.Graph()
    g.add_simple_conditional("name", "condition", "value_true", "value_false")
    g.add_step_quick("condition", lambda pre_req: pre_req)
    g.finalize_definition()

    # Here, condition and possibilities are unreachable
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    # Here, condition is undefined but reachable, but all values are unreachable
    g["pre_req"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    # Now one of the possibilities is already available, therefore the conditional might be, depending on condition
    g["value_true"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "uncertain"
    g.clear_reachabilities()

    # Now all of the possibilities are already available, therefore the conditional is certainly reachable
    g["value_false"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "reachable"


def test_multiple_conditional_no_conditions_defined():
    g = gr.Graph()
    g.add_multiple_conditional("name", ["ca", "cb"], ["va", "vb", "vc"])
    g.add_step_quick("ca", lambda pa: pa)
    g.add_step_quick("cb", lambda pb: pb)
    g.finalize_definition()

    # Here, condition and possibilities are unreachable
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    # Here, ca is undefined but reachable, but all values are unreachable
    g["pa"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "unreachable"
    g.clear_reachabilities()

    # Now one of the possibilities is already available, therefore the conditional might be, depending on condition
    g["va"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "uncertain"
    g.clear_reachabilities()

    # Now all of the possibilities are reachable, but the conditional is still uncertain because we do not know which condition is True
    g["pb"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "uncertain"
    g.clear_reachabilities()

    # Now all of the possibilities are already available, therefore the conditional is certainly reachable
    g["vb"] = 1
    g["vc"] = 1
    g.find_reachability_targets("name")
    assert g.get_reachability("name") == "reachable"


def test_execution_with_feasibility_check():
    g = gr.Graph()
    g.add_step("b", "fb", "a")
    g["fb"] = lambda a: a
    g.finalize_definition()

    # a is not available
    with pytest.raises(ValueError):
        gr.execute_graph_from_context(g, {}, "b")
    # Now a becomes available
    res = gr.execute_graph_from_context(g, {"a": 1}, "b")
    assert res["b"] == 1


def test_execution_with_feasibility_check_uncertain():
    g = gr.Graph()
    g.add_simple_conditional("name", "condition", "value_true", "value_false")
    g.add_step_quick("condition", lambda pre_req: pre_req)
    g["pre_req"] = True
    g["value_true"] = 1
    g.finalize_definition()

    # Reachability is uncertain because condition is reachable and one value also is
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        res = gr.execute_graph_from_context(g, {}, "name")
        assert len(w) == 1
        # But the computation is actually feasible
        assert res["name"] == 1


def test_unfeasible_wrap():
    g = gr.Graph()
    g.add_step("d", "op_d", "a", "b", "c")
    g["op_d"] = lambda a, b, c: a + b + c
    g.finalize_definition()

    with pytest.raises(ValueError):
        # Pass b as input, c as constant, but do not pass a
        f1 = gr.wrap_graph_with_function(g, ["b"], "d", constants={"c": 1}, input_as_kwargs=False)
    # Pass also a
    f2 = gr.wrap_graph_with_function(g, ["a", "b"], "d", constants={"c": 1}, input_as_kwargs=False)
    assert f2(1, 1) == 3



"""
Core of the grapes package. Includes the classes for nodes and graphs.

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

from typing import ValuesView
import networkx as nx
from . import function_composer
import inspect


starting_node_properties = {"type": "standard", "has_value": False, "value": None, "is_frozen": False, "is_recipe": False, "topological_generation_index": -1, "has_reachability": False, "reachability": None}


class Graph():
    """
    Class that represents a graph of nodes.
    """

    def __init__(self, nx_digraph=None):
        # Internally, we handle a nx_digraph
        if nx_digraph == None:
            self._nxdg = nx.DiGraph()
        else:
            self._nxdg = nx_digraph
        # Alias for easy access
        self.nodes = self._nxdg.nodes

    def __getitem__(self, node):
        """
        Get the value of a node with []
        """
        return self.get_value(node)

    def __setitem__(self, node, value):
        """
        Set the value of a node with []
        """
        self.set_value(node, value)

    def __eq__(self, other):
        """
        Equality check based on all members.
        """
        return (isinstance(other, self.__class__) and nx.is_isomorphic(self._nxdg, other._nxdg, dict.__eq__, dict.__eq__))

    def add_step(self, name, recipe=None, *args, **kwargs):
        """
        Interface to add a node to the graph, with all its dependencies.
        """
        # Check that if a node has dependencies, it also has a recipe
        if recipe is None and (len(args) > 0 or len(kwargs.keys()) > 0):
            raise ValueError("Cannot add node with dependencies without a recipe")

        elif recipe is None:  # Accept nodes with no dependencies
            # The node is readded even if already existing
            self._nxdg.add_node(name, **starting_node_properties)

        else:  # Standard case
            # Add the node
            # The node is readded even if already existing
            self._nxdg.add_node(name, **starting_node_properties)
            # Set attributes
            # Note: This could be done in the constructor, but doing it separately adds flexibility
            # Indeed, we might want to change how attributes work, and we can do it by modifying setters
            self.set_recipe(name, recipe)
            self.set_args(name, args)
            self.set_kwargs(name, kwargs)

            # Add and connect the recipe
            # Avoid adding existing recipe so as not to overwrite attributes
            if recipe not in self.nodes:
                self._nxdg.add_node(recipe, **starting_node_properties)
            self.set_is_recipe(recipe, True)
            # Note: adding argument to the edges is elegant but impractical.
            # If relations were defined through edges attributes rather than stored inside nodes,
            # retrieving them would require iterating through all edges and selecting the ones with the right attributes.
            # Although feasible, this is much slower than simply accessing node attributes.
            self._nxdg.add_edge(recipe, name)

            # Add and connect the other dependencies
            for arg in args:
                # Avoid adding existing dependencies so as not to overwrite attributes
                if arg not in self.nodes:
                    self._nxdg.add_node(arg, **starting_node_properties)
                self._nxdg.add_edge(arg, name)
            for value in kwargs.values():
                # Avoid adding existing dependencies so as not to overwrite attributes
                if value not in self.nodes:
                    self._nxdg.add_node(value, **starting_node_properties)
                self._nxdg.add_edge(value, name)

    def add_step_quick(self, name, recipe):
        """
        Interface to quickly add a step by passing a name and a function.

        The recipe node takes the name of the passed function.
        Dependency nodes are built from the args and kwonlyargs of the passed function.
        """
        # Check that the passed recipe is a valid function
        if not inspect.isfunction(recipe):
            raise TypeError("The passed recipe should be a function, but it is a " + type(recipe))
        argspec = inspect.getfullargspec(recipe)
        # varargs and varkw are not supported because add_step_quick needs parameter names to build nodes
        if argspec.varargs is not None or argspec.varkw is not None:
            raise ValueError("Functions with varargs or varkwargs are not supported by add_step_quick because there would be no way to name dependency nodes")

        # Get function name and parameters
        recipe_name = recipe.__name__
        # Lambdas are all automatically named "<lambda>" so we change this
        if recipe_name == "<lambda>":
            recipe_name = "recipe_for_" + name
        args = argspec.args
        kwargs_list = argspec.kwonlyargs
        # Build a dictionary with identical keys and values so that recipe is called all the keys are used are kwargs
        kwargs = {kw: kw for kw in kwargs_list}
        # Add the step: this will create nodes for name, recipe_name and all elements of args and kwargs_list
        self.add_step(name, recipe_name, *args, **kwargs)
        # Directly set the value of recipe_name to recipe
        self.set_value(recipe_name, recipe)

    def add_simple_conditional(self, name, condition, value_true, value_false):
        """
        Interface to add a conditional to the graph.
        """
        # Add all nodes
        self._nxdg.add_node(name, **starting_node_properties)
        for node in [condition, value_true, value_false]:
            # Avoid adding existing dependencies so as not to overwrite attributes
            if node not in self.nodes:
                self._nxdg.add_node(node, **starting_node_properties)

        # Connect edges
        self._nxdg.add_edge(condition, name)
        self._nxdg.add_edge(value_true, name)
        self._nxdg.add_edge(value_false, name)

        # Specify that this node is a conditional
        self.set_type(name, "conditional")

        # Add conditions name to the list of conditions of the conditional
        self.set_conditions(name, [condition])

        # Add possibilities to the list of possibilities of the conditional
        self.set_possibilities(name, [value_true, value_false])

    def add_multiple_conditional(self, name, conditions, possibilities):
        """
        Interface to add a multiple conditional to the graph.
        """
        # Add all nodes and connect all edges
        self._nxdg.add_node(name, **starting_node_properties)
        for node in conditions + possibilities:
            # Avoid adding existing dependencies so as not to overwrite attributes
            if node not in self.nodes:
                self._nxdg.add_node(node, **starting_node_properties)
            self._nxdg.add_edge(node, name)

        # Specify that this node is a conditional
        self.set_type(name, "conditional")

        # Add conditions name to the list of conditions of the conditional
        self.set_conditions(name, conditions)

        # Add possibilities to the list of possibilities of the conditional
        self.set_possibilities(name, possibilities)

    def edit_step(self, name, recipe=None, *args, **kwargs):
        """
        Interface to edit an existing node, changing its predecessors
        """
        if name not in self.nodes:
            raise ValueError("Cannot edit non-existent node " + name)

        # Store old attributes
        was_recipe = self.is_recipe(name)
        was_frozen = self.is_frozen(name)
        had_value = self.has_value(name)
        if had_value:
            old_value = self.get_value(name)

        # Remove in-edges from the node because we need to replace them
        # use of list() is to make a copy because in_edges() returns a view
        self._nxdg.remove_edges_from(list(self._nxdg.in_edges(name)))
        # Readd the step. This should not break anything
        self.add_step(name, recipe, *args, **kwargs)

        # Readd attributes
        # Readding out-edges is not needed because we never removed them
        self.set_is_recipe(name, was_recipe)
        self.set_is_frozen(name, was_frozen)
        self.set_has_value(name, had_value)
        if had_value:
            self.set_value(name, old_value)

    def remove_step(self, name):
        """
        Interface to remove an existing node, without changing anything else
        """
        if name not in self.nodes:
            raise ValueError("Cannot edit non-existent node " + name)
        self._nxdg.remove_node(name)

    def get_node_attribute(self, node, attribute):
        attributes = self.nodes[node]
        if attribute in attributes and attributes[attribute] is not None:
            return attributes[attribute]
        else:
            raise ValueError("Node " + node + " has no " + attribute)

    def set_node_attribute(self, node, attribute, value):
        self.nodes[node][attribute] = value

    def is_recipe(self, node):
        return self.get_node_attribute(node, "is_recipe")

    def set_is_recipe(self, node, is_recipe):
        return self.set_node_attribute(node, "is_recipe", is_recipe)

    def get_recipe(self, node):
        return self.get_node_attribute(node, "recipe")

    def set_recipe(self, node, recipe):
        return self.set_node_attribute(node, "recipe", recipe)

    def get_args(self, node):
        return self.get_node_attribute(node, "args")

    def set_args(self, node, args):
        return self.set_node_attribute(node, "args", args)

    def get_kwargs(self, node):
        return self.get_node_attribute(node, "kwargs")

    def set_kwargs(self, node, kwargs):
        return self.set_node_attribute(node, "kwargs", kwargs)

    def get_conditions(self, node):
        conditions = self.get_node_attribute(node, "conditions")
        if not isinstance(conditions, list):
            conditions = list(conditions)
        return conditions

    def set_conditions(self, node, conditions):
        if not isinstance(conditions, list):
            conditions = list(conditions)
        return self.set_node_attribute(node, "conditions", conditions)

    def get_possibilities(self, node):
        possibilities = self.get_node_attribute(node, "possibilities")
        if not isinstance(possibilities, list):
            possibilities = list(possibilities)
        return possibilities

    def set_possibilities(self, node, possibilities):
        if not isinstance(possibilities, list):
            possibilities = list(possibilities)
        return self.set_node_attribute(node, "possibilities", possibilities)

    def get_type(self, node):
        return self.get_node_attribute(node, "type")

    def set_type(self, node, type):
        return self.set_node_attribute(node, "type", type)

    def get_topological_generation_index(self, node):
        return self.get_node_attribute(node, "topological_generation_index")

    def set_topological_generation_index(self, node, index):
        self.set_node_attribute(node, "topological_generation_index", index)

    def get_value(self, node):
        attributes = self.nodes[node]
        if "value" in attributes and attributes["value"] is not None and self.nodes[node]["has_value"]:
            return attributes["value"]
        else:
            raise ValueError("Node " + node + " has no value")

    def set_value(self, node, value):
        # Note: This changes reachability
        self.nodes[node]["value"] = value
        self.nodes[node]["has_value"] = True

    def unset_value(self, node):
        # Note: This changes reachability
        self.nodes[node]["has_value"] = False

    def get_reachability(self, node):
        attributes = self.nodes[node]
        if "reachability" in attributes and attributes["reachability"] is not None and self.nodes[node]["has_reachability"]:
            return attributes["reachability"]
        else:
            raise ValueError("Node " + node + " has no reachability")

    def set_reachability(self, node, reachability):
        if reachability not in ("unreachable", "uncertain", "reachable"):
            raise ValueError(reachability + " is not a valid reachability value.")
        self.nodes[node]["reachability"] = reachability
        self.nodes[node]["has_reachability"] = True

    def unset_reachability(self, node):
        self.nodes[node]["has_reachability"] = False

    def is_frozen(self, node):
        return self.get_node_attribute(node, "is_frozen")

    def set_is_frozen(self, node, is_frozen):
        return self.set_node_attribute(node, "is_frozen", is_frozen)

    def has_value(self, node):
        return self.get_node_attribute(node, "has_value")

    def set_has_value(self, node, has_value):
        return self.set_node_attribute(node, "has_value", has_value)

    def clear_values(self, *args):
        """
        Clear values in the graph nodes.
        """
        if len(args) == 0:  # Interpret as "Clear everything"
            nodes_to_clear = self.nodes
        else:
            nodes_to_clear = args & self.nodes  # Intersection

        for node in nodes_to_clear:
            if self.is_frozen(node):
                continue
            self.unset_value(node)

    def has_reachability(self, node):
        return self.get_node_attribute(node, "has_reachability")

    def set_has_reachability(self, node, has_reachability):
        return self.set_node_attribute(node, "has_reachability", has_reachability)

    def clear_reachabilities(self, *args):
        """
        Clear reachabilities in the graph nodes.
        """
        if len(args) == 0:  # Interpret as "Clear everything"
            nodes_to_clear = self.nodes
        else:
            nodes_to_clear = args & self.nodes  # Intersection

        for node in nodes_to_clear:
            if self.is_frozen(node):
                continue
            self.unset_reachability(node)

    def update_internal_context(self, dictionary):
        """
        Update internal context with a dictionary.

        Parameters
        ----------
        dictionary: dict
            Dictionary with the new values
        """
        for key, value in dictionary.items():
            # Accept dictionaries with more keys than needed
            if key in self.nodes:
                self.set_value(key, value)

    def set_internal_context(self, dictionary):
        """
        Clear all values and then set a new internal context with a dictionary.

        Parameters
        ----------
        dictionary: dict
            Dictionary with the new values
        """
        self.clear_values()
        self.update_internal_context(dictionary)

    def get_internal_context(self, exclude_recipes=False):
        """
        Get the internal context.

        Parameters
        ----------
        exclude_recipes: bool
            Whether to exclude recipes from the returned dictionary or keep them.
        """
        if exclude_recipes:
            return {key: self.get_value(key) for key in self.nodes if (self.has_value(key) and not self.is_recipe(key))}
        else:
            return {key: self.get_value(key) for key in self.nodes if self.has_value(key)}

    def get_list_of_values(self, list_of_keys):
        """
        Get values as list.

        Parameters
        ----------
        list_of_keys: list of hashables (typically strings)
            List of names of nodes whose values are required

        Returns
        -------
        list
            List like list_of_keys which contains values of nodes
        """
        res = []
        for key in list_of_keys:
            res.append(self.get_value(key))
        return res

    def get_dict_of_values(self, list_of_keys):
        """
        Get values as dictionary.

        Parameters
        ----------
        list_of_keys: list of hashables (typically strings)
            List of names of nodes whose values are required

        Returns
        -------
        dict
            Dictionary whose keys are the elements of list_of_keys and whose values are the corresponding node values
        """
        return {key: self.get_value(key) for key in list_of_keys}

    def get_kwargs_values(self, dictionary):
        """
        Get values from the graph, using a dictionary that works like function kwargs.

        Parameters
        ----------
        dictionary: dict
            Keys in dictionary are to be interpreted as keys for function kwargs, while values in dictionary are node names

        Returns
        -------
        dict
            A dict with the same keys of the input dictionary, but with values replaced by the values of the nodes
        """
        return {key: self.get_value(value) for key, value in dictionary.items()}

    def evaluate_target(self, target, continue_on_fail=False):
        """
        Generic interface to evaluate a GenericNode.
        """
        if self.get_type(target) == "standard":
            return self.evaluate_standard(target, continue_on_fail)
        elif self.get_type(target) == "conditional":
            return self.evaluate_conditional(target, continue_on_fail)
        else:
            raise ValueError("Evaluation of nodes of type " + self.get_type(target) + " is not supported")

    def evaluate_standard(self, node, continue_on_fail=False):
        """
        Evaluate of a node.
        """
        # Check if it already has a value
        if self.has_value(node):
            self.get_value(node)
            return
        # If not, evaluate all arguments
        for dependency_name in self._nxdg.predecessors(node):
            self.evaluate_target(dependency_name, continue_on_fail)

        # Actual computation happens here
        try:
            recipe = self.get_recipe(node)
            func = self.get_value(recipe)
            res = func(*self.get_list_of_values(self.get_args(node)), **self.get_kwargs_values(self.get_kwargs(node)))
        except Exception as e:
            if continue_on_fail:
                # Do nothing, we want to keep going
                return
            else:
                if len(e.args) > 0:
                    e.args = ("While evaluating " + node + ": " + str(e.args[0]),) + e.args[1:]
                raise
        # Save results
        self.set_value(node, res)

    def evaluate_conditional(self, conditional, continue_on_fail=False):
        """
        Evaluate a conditional.
        """
        # Check if it already has a value
        if self.has_value(conditional):
            self.get_value(conditional)
            return
        # If not, evaluate the conditions until one is found true
        for index, condition in enumerate(self.get_conditions(conditional)):
            self.evaluate_target(condition, continue_on_fail)
            if self.has_value(condition) and self.get_value(condition):
                break
            elif not self.has_value(condition):
                # Computing failed
                if continue_on_fail:
                    # Do nothing, we want to keep going
                    return
                else:
                    raise ValueError("Node " + condition + " could not be computed")
        else:  # Happens if loop is never broken, i.e. when no conditions are true
            index = -1

        # Actual computation happens here
        try:
            possibility = self.get_possibilities(conditional)[index]
            self.evaluate_target(possibility, continue_on_fail)
            res = self.get_value(possibility)
        except:
            if continue_on_fail:
                # Do nothing, we want to keep going
                return
            else:
                raise ValueError("Node " + possibility + " could not be computed")
        # Save results and release
        self.set_value(conditional, res)

    def execute_to_targets(self, *targets):
        """
        Evaluate all nodes in the graph that are needed to reach the targets.
        """
        for target in targets:
            self.evaluate_target(target, False)

    def progress_towards_targets(self, *targets):
        """
        Move towards the targets by evaluating nodes, but keep going if evaluation fails.
        """
        for target in targets:
            self.evaluate_target(target, True)

    def find_reachability_target(self, target):
        """
        Generic interface to find the reachability of a GenericNode.
        """
        if self.get_type(target) == "standard":
            return self.find_reachability_standard(target)
        elif self.get_type(target) == "conditional":
            return self.find_reachability_conditional(target)
        else:
            raise ValueError("Finding the reachability of nodes of type " + self.get_type(target) + " is not supported")

    def find_reachability_standard(self, node):
        """
        Find the reachability of a standard node.
        """
        # Check if it already has a reachability
        if self.has_reachability(node):
            return
        # Check if it already has a value
        if self.has_value(node):
            self.set_reachability(node, "reachable")
            return
        # If not, check the missing dependencies of all arguments
        dependencies = set(self._nxdg.predecessors(node))
        if len(dependencies) == 0:
            # If this node does not have predecessors (and does not have a value itself), it is not reachable
            self.set_reachability(node, "unreachable")
            return
        # Otherwise, dependencies must be checked
        self.find_reachability_targets(*dependencies)
        self.set_reachability(node, self.get_worst_reachability(*dependencies))

    def find_reachability_conditional(self, conditional):
        """
        Find the reachability of a conditional.
        """
        # Check if it already has a reachability
        if self.has_reachability(conditional):
            return
        # Check if it already has a value
        if self.has_value(conditional):
            self.get_value(conditional)
            self.set_reachability(conditional, "reachable")
            return
        # If not, evaluate the conditions until one is found true
        for index, condition in enumerate(self.get_conditions(conditional)):
            if self.has_value(condition) and self.get_value(condition):
                # A condition is true
                possibility = self.get_possibilities(conditional)[index]
                self.find_reachability_target(possibility)
                self.set_reachability(conditional, self.get_reachability(possibility))
                return
        else:
            # Happens if loop is never broken, i.e. when no conditions are true
            # If all conditions and possibilities are reachable -> reachable
            # If all conditions and possibilities are unreachable -> unreachable
            # If some conditions are reachable or uncertain but the corresponding possibilities are all unreachable -> unreachable
            # In all other cases -> uncertain
            self.find_reachability_targets(*self.get_conditions(conditional))
            self.find_reachability_targets(*self.get_possibilities(conditional))

            if self.get_worst_reachability(*(self.get_conditions(conditional)+self.get_possibilities(conditional))) == "reachable":
                # All conditions and possibilities are reachable -> reachable
                self.set_reachability(conditional, "reachable")
            elif self.get_best_reachability(*(self.get_conditions(conditional)+self.get_possibilities(conditional))) == "unreachable":
                # All conditions and possibilities are unreachable -> unreachable
                self.set_reachability(conditional, "unreachable")
            else:
                not_unreachable_condition_possibilities = []
                for index, condition in enumerate(self.get_conditions(conditional)):
                    if self.get_reachability(condition) != "unreachable":
                        not_unreachable_condition_possibilities.append(self.get_possibilities(conditional)[index])
                if self.get_best_reachability(*not_unreachable_condition_possibilities) == "unreachable":
                    # All corresponding possibilities are unreachable -> unreachable
                    self.set_reachability(conditional, "unreachable")
                else:
                    self.set_reachability(conditional, "uncertain")

    def get_worst_reachability(self, *nodes):
        list_of_reachabilities = []
        for node in nodes:
            list_of_reachabilities.append(self.get_reachability(node))
        if "unreachable" in list_of_reachabilities:
            return "unreachable"
        elif "uncertain" in list_of_reachabilities:
            return "uncertain"
        else:
            return "reachable"

    def get_best_reachability(self, *nodes):
        list_of_reachabilities = []
        for node in nodes:
            list_of_reachabilities.append(self.get_reachability(node))
        if "reachable" in list_of_reachabilities:
            return "reachable"
        elif "uncertain" in list_of_reachabilities:
            return "uncertain"
        else:
            return "unreachable"

    def find_reachability_targets(self, *targets):
        for target in targets:
            self.find_reachability_target(target)

    def is_other_node_compatible(self, node, other, other_node):
        # If types differ, return False
        if self.get_type(node) != other.get_type(other_node):
            return False
        # If nodes are equal, return True
        if self.nodes[node] == other._nxdg.nodes[other_node]:
            return True
        # If they both have values but they differ, return False. If only one has a value, proceed
        if self.has_value(node) and other.has_value(other_node) and self.get_value(node) != other.get_value(other_node):
            # Plot twist! Both are functions and have the same code: proceed
            if inspect.isfunction(self.get_value(node)) and inspect.isfunction(other.get_value(other_node)) and self.get_value(node).__code__.co_code == other.get_value(other_node).__code__.co_code:
                pass
            else:
                return False
        # If they both have dependencies but they differ, return False. If only one has dependencies, proceed
        predecessors = list(self._nxdg.predecessors(node))
        other_predecessors = list(other._nxdg.predecessors(other_node))
        if len(predecessors) != 0 and len(other_predecessors) != 0 and predecessors != other_predecessors:
            return False
        # Return True if at least one has no dependencies (or they are the same), at least one has no value (or they are the same)
        return True

    def is_compatible(self, other):
        """
        Check if self and other can be composed. Currently DAG status is not verified.
        """
        if not isinstance(other, Graph):
            return False
        common_nodes = self.nodes & other._nxdg.nodes  # Intersection
        for key in common_nodes:
            if not self.is_other_node_compatible(key, other, key):
                return False
        return True

    def merge(self, other):
        """
        Merge other into self.
        """
        if not self.is_compatible(other):
            raise ValueError("Cannot merge incompatible graphs")
        res = nx.compose(self._nxdg, other._nxdg)
        self._nxdg = res
        # Refresh alias for easy access
        self.nodes = self._nxdg.nodes

    def simplify_dependency(self, node_name, dependency_name):
        # Make everything a keyword argument. This is the fate of a simplified node
        self.get_kwargs(node_name).update({argument: argument for argument in self.get_args(node_name)})
        # Build lists of dependencies
        func_dependencies = list(self.get_kwargs(node_name).values())
        subfuncs = []
        subfuncs_dependencies = []
        for argument in self.get_kwargs(node_name):
            if argument == dependency_name:
                if self.get_type(dependency_name) != "standard":
                    raise TypeError("Simplification only supports standard nodes, while the type of " + dependency_name + " is " + self.get_type(dependency_name))
                if self.get_type(self.get_recipe(dependency_name)) != "standard":
                    raise TypeError("Simplification only supports standard nodes, while the type of " + self.get_recipe(dependency_name) + " is " + self.get_type(self.get_recipe(dependency_name)))
                subfuncs.append(self[self.get_recipe(dependency_name)])  # Get python function
                subfuncs_dependencies.append(list(self.get_args(dependency_name)) + list(self.get_kwargs(dependency_name).values()))
            else:
                subfuncs.append(function_composer.identity_token)
                subfuncs_dependencies.append([argument])
        # Compose the functions
        self[self.get_recipe(node_name)] = function_composer.function_compose_simple(self[self.get_recipe(node_name)], subfuncs, func_dependencies, subfuncs_dependencies)
        # Change edges
        self._nxdg.remove_edge(dependency_name, node_name)
        for argument in self.get_args(dependency_name) + tuple(self.get_kwargs(dependency_name).values()):
            self._nxdg.add_edge(argument, node_name, accessor=argument)
        # Update node
        self.set_args(node_name, ())
        new_kwargs = self.get_kwargs(node_name)
        new_kwargs.update({argument: argument for argument in self.get_args(dependency_name) + tuple(self.get_kwargs(dependency_name).values())})
        new_kwargs = {key: value for key, value in new_kwargs.items() if value != dependency_name}
        self.set_kwargs(node_name, new_kwargs)

    def simplify_all_dependencies(self, node_name, exclude=[]):
        dependencies = self.get_args(node_name) + tuple(self.get_kwargs(node_name).values())
        for dependency in dependencies:
            if dependency not in exclude:
                self.simplify_dependency(node_name, dependency)

    def freeze(self, *args):
        if len(args) == 0:  # Interpret as "Freeze everything"
            nodes_to_freeze = self.nodes
        else:
            nodes_to_freeze = args & self.nodes  # Intersection

        for key in nodes_to_freeze:
            if self.has_value(key):
                self.set_is_frozen(key, True)

    def unfreeze(self, *args):
        if len(args) == 0:  # Interpret as "Unfreeze everything"
            nodes_to_unfreeze = self.nodes.keys()
        else:
            nodes_to_unfreeze = args & self.nodes  # Intersection

        for key in nodes_to_unfreeze:
            self.set_is_frozen(key, False)

    def make_recipe_dependencies_also_recipes(self):
        """
        Make dependencies (parents) of recipes also recipes
        """
        # Work in reverse topological order, to get children before parents
        for node in reversed(self.get_topological_order()):
            if self.is_recipe(node):
                for parent in self._nxdg.predecessors(node):
                    self.set_is_recipe(parent, True)

    def finalize_definition(self):
        """
        Perform operations that should typically be done after the definition of a graph is completed

        Currently, this freezes all values, because it is assumed that values given during definition are to be frozen.
        It also marks dependencies of recipes as recipes themselves.
        """
        self.make_recipe_dependencies_also_recipes()
        self.update_topological_generation_indexes()
        self.freeze()

    def get_topological_order(self):
        """
        Return list of nodes in topological order, i.e., from dependencies to targets
        """
        return list(nx.topological_sort(self._nxdg))

    def get_topological_generations(self):
        """
        Return list of topological generations of the graph
        """
        return list(nx.topological_generations(self._nxdg))

    def update_topological_generation_indexes(self):
        generations = self.get_topological_generations()
        for node in self.nodes:
            for index, generation in enumerate(generations):
                if node in generation:
                    self.set_topological_generation_index(node, index)
                    break



"""
Setup file to install grapes as a package.

Usage: pip install .

Author: Giulio Foletto <giulio.foletto@outlook.com>.
License: See project-level license file.
"""

from setuptools import setup

# Assign the content of README.md to long_description
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='grapes',
      version='0.6.0',
      description='Helper for dataflow based programming',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/giuliofoletto/grapes',
      author='Giulio Foletto',
      author_email='giulio.foletto@outlook.com',
      license='Apache',
      license_files=('LICENSE.txt', 'NOTICE.txt'),
      packages=['grapes'],
      install_requires=[
          'networkx',
      ],
      extras_require={
          'visualize': ['pygraphviz', 'matplotlib'],  # For installation of pygraphviz, refer to https://pygraphviz.github.io/documentation/stable/install.html
          'testing': ['pytest']
      },
      zip_safe=False)
